#ifndef MODELINGFD_AUX_H
#define MODELINGFD_AUX_H 1

#include "hypercube.h"

namespace Yang {

class wavFieldGroup{
public:
   cubeData<ETYPE>      p_cb[3]; //acoustic wavefield pressure, cube data
   cubeData<ETYPE>    * pp_cb[3];//pointers for array p
private:
   wavFieldGroup( const wavFieldGroup& wfg);
   
public: //method
   wavFieldGroup() {
	  pp_cb[0] = NULL;
	  pp_cb[1] = NULL;
	  pp_cb[2] = NULL;
   }
   void switchWavfieldPtrs();
   void getPtrs(cubeData<ETYPE> **ppp0, cubeData<ETYPE> **ppp1, cubeData<ETYPE> **ppp2);
   void reset();
   void alloc(const hypercube& shape);
   void dealloc();
};

inline void wavFieldGroup::switchWavfieldPtrs()
{
   cubeData<ETYPE> *p_tmp;
   p_tmp = pp_cb[0];
   pp_cb[0] = pp_cb[1];
   pp_cb[1] = pp_cb[2];
   pp_cb[2] = p_tmp;
}

inline void wavFieldGroup::reset()
{
   pp_cb[0]->zero();
   pp_cb[1]->zero();
   pp_cb[2]->zero();
}

inline void wavFieldGroup::alloc(const hypercube& shape)
{
   p_cb[0].alloc(shape);
   p_cb[1].alloc(shape);
   p_cb[2].alloc(shape);   
   pp_cb[0] = &(p_cb[0]);
   pp_cb[1] = &(p_cb[1]);
   pp_cb[2] = &(p_cb[2]);
}

inline void wavFieldGroup::dealloc()
{
   p_cb[0].dealloc();
   p_cb[1].dealloc();
   p_cb[2].dealloc();
   pp_cb[0] = NULL;
   pp_cb[1] = NULL;
   pp_cb[2] = NULL;   
}

//cannotmake virtual because the argument type is different from 3D case
inline void wavFieldGroup::getPtrs(cubeData<ETYPE> **ppp0, cubeData<ETYPE> **ppp1, cubeData<ETYPE> **ppp2)
{
   *ppp0 = pp_cb[0];
   *ppp1 = pp_cb[1];
   *ppp2 = pp_cb[2];
}

//Params of doing prestack imaging
//Measured by grid points in wavfield grid pts.
struct offsetImagingParams {
   bool bDoOffsetGather;
   int ohx,dhx,nhx;
   int ohy,dhy,nhy;
   int ohz,dhz,nhz;
	int ot,dt,nt;
   unsigned int AxesUsage; //Use the least siginifcant bits to indicate the which axis/axes(hx,hy,hz,t) is used
   offsetImagingParams(): bDoOffsetGather(false),ohx(0),dhx(0),nhx(0), ohy(0),dhy(0),nhy(0),
	  ohz(0),dhz(0),nhz(0), AxesUsage(0) {}
};

//The axes are (x,y,z,hx [,hy] [,hz], [t])
//Only deal with sptial axes currently
//If input wfld, then output wfld shape slices //For Image output
//If input vel grid, then ouptut vel grid shape slices.//For image input
inline void get_prestack_img_cube_shape(hypercube& newShape, offsetImagingParams& off_img_param, const hypercube& wfldShape, float dtComp = 0.0f) {
   assert(off_img_param.bDoOffsetGather);
   vecAxis axess = wfldShape.get_axes();
   unsigned int axesUsage = off_img_param.AxesUsage;
   axis ax, ax_s; float d;
   if (axesUsage & 0x1) {//Use hx axis, remember that x axis is axis 2(TWO) in wfldShape
	  ax.label = string("hx");
	  ax_s = wfldShape.get_axis(2); d=ax_s.d;
	  ax.o = off_img_param.ohx * d;
	  ax.d = ax_s.d * off_img_param.dhx;
	  ax.n = off_img_param.nhx;
	  axess.push_back(ax);
   }
   if (axesUsage & 0x2) {//Use hy axis
	  ax.label = string("hy");
	  ax_s = wfldShape.get_axis(3); d=ax_s.d;
	  ax.o = off_img_param.ohy * d;
	  ax.d = ax_s.d * off_img_param.dhy;
	  ax.n = off_img_param.nhy;
	  axess.push_back(ax);
   }
   if (axesUsage & 0x4) {//Use hz axis
	  ax.label = string("hz");
	  ax_s = wfldShape.get_axis(1); d=ax_s.d;
	  ax.o = off_img_param.ohz * d;
	  ax.d = ax_s.d * off_img_param.dhz;
	  ax.n = off_img_param.nhz;
	  axess.push_back(ax);
   }      
	newShape.set_axes(axess);
	return;
}

enum impose_receiver_record_type
{
   AsForce,
   AsBoundary
};
enum boundaryType 
{
   abs_null, //do nothing
   abs_ord1, //absorb 1st order
   abs_ord2, //absorb 2nd order
   abs_Dave,  //Improved 1st order absorption method by Dave Hale, used in SU model
   abs_ord1par, //absorb 1st order, but only first order accuracy in FD scheme, but easy to parallel.
};

enum action_type {
   acModeling,
   acMigration,
   acBornModeling,
   acTomFwd, //From deltaS -> delta I
   acTomAdj  //From deltaI -> delta S
};

class vel;
class srcWav;

//Wave propagate params
struct propagateParams {
  //3D field for wavefield snap could be an out of core object or incore.
  dmRsf<float> *pSrcSnapshot;
  dmRsf<float> *pRecvSnapshot;
  
  dmRsf<float> *pSrcSnapshot2;
  dmRsf<float> *pRecvSnapshot2;
  
  //FOR SIMPLICITY, ALL QUANTITIES ARE OF WAVEFIELD SIZE
  //All For zero offset
  dmRsf<float> *pImages; //stores RTM_image every shot, for zero offset
  cubeData<float> *pFinalImage; //Image summing up all shots, also as the input refl model for adjRTM
  
  cubeData<float> *pFinalDeltaS;  //This is always incore, input for born Modeling, for zero-offset
  cubeData<float>*  pDeltaSs; //offset Image gathers summing all shots.
  
  //For prestack image with subsurface offset, for simplicity, these quantities are all of wavefield size grids.
  offsetImagingParams mOffImgParams;
  dmRsf<float>*		pOffImages; 	//offset Image gathers for each shot
  cubeData<float>*  pFinalOffImage;//summing all shots
  
  
  //the seismic record, input for Imaging, output of modeling
  dmRsf<float> *pRsfSeisRec; 	 
  int ndim; //2 or 3, propagate domain dimension
  //Only this velocity quantity keep the size of velocity.
  vel* pVel;
  srcWav* pWavelet;
  
  //Src Receiver Information. Needs to be further improved to accommodate generalization
  //Source Receiver Geometry part  //Get shot info, here use regular geometry

  bool b3D;
	int m_nShots  ,m_nTraces; //number of sources and receivers along the inline direction
  int m_nShotsY ,m_nTracesY; //number of sources and receivers cross-line direction
  
	int mShot1_ix,mShot1_iz,mShot1_iy,mShotIncX,mShotIncY;
	int mTrace1ix,mTrace1iz,mTrace1iy,mTraceIncX,mTraceIncY;
	bool bShot_offset; //yes means marine acquisition, no means land acquisition.
  
	//For snap/Dump params
	int   j_shot_snap, f_shot_snap, n_shot_snap;
	bool  bSaveWfldSnap;
	
	impose_receiver_record_type mImposeRecvType;
	
	//velocity model grid shape, global vel Model
	hypercube velModelShape;  //The entire vel model
	hypercube finalImageShape;  //The entire image
  hypercube finalOffImageShape;  //The entire offset image (if there is).
  
	//For each shot
 	hypercube modelGridShape; //vel model grid size 
	hypercube mShapeWavfield; //The dimension of wavefield
  hypercube mImageShape;    //The size of image per shot, HERE I remove the boundary and the padding zone
  
  bool bDoSrcInt; //If do src intensity or not.
  bool bDoRecvInt;
public:
  propagateParams():m_nShots(1),m_nShotsY(1),j_shot_snap(0),f_shot_snap(0),n_shot_snap(0),
    bSaveWfldSnap(false),bDoSrcInt(false), bDoRecvInt(false) {
    //Put default values here.
    pSrcSnapshot 	= NULL;
    pRecvSnapshot	= NULL;
    pSrcSnapshot2 	= NULL;
    pRecvSnapshot2	= NULL;
    pImages		= pOffImages = NULL;
    pFinalImage  = pFinalOffImage = NULL;
    pDeltaSs = pFinalDeltaS = NULL;
    pVel = NULL;
    pWavelet = NULL;
  }
    
  //Make copy constructors
  propagateParams(const propagateParams& other);
};


inline void checkSnapRecordParams(propagateParams* par) {
	int o,j;
	o=par->f_shot_snap;   j=par->j_shot_snap;
	VERIFYC((o) >= 0);
	VERIFYC((j) >= 1);
	if (par->n_shot_snap == 0)	{ //record all shots
		par->n_shot_snap = (par->m_nShots-1-o)/j + 1;
	} else {
		VERIFYC(par->n_shot_snap >= 1);
		VERIFYC(par->f_shot_snap + (par->n_shot_snap-1)*(par->j_shot_snap) < par->m_nShots);
	}
}

struct FDModelingTimeParams {
   float totalTime;
   float dtComp,dtRecord;
   float dtSnap,dtImaging;
   
   int Nt,nSnap,nImaging,nRec;
   int jSnap,jRec,jImaging;
  
   //default copy constructor is OK, no pointers
};

//For compatibility, this structure contain both oneway and two-way parameters.
struct FDModelingParams: public propagateParams{
   
  FDModelingTimeParams timeParam;
	int stencilWidth;
	bool bFreeSurfaceOnTop;
	bool bUse4thOrder;
		
   int offx,offz,offy; //off between vel model cube and wavefield cube
	//padding layer related
	int m_nPad;
	//boundary layer
	boundaryType m_bdType;
	
	bool bTake2ndTimeDerivForRTM;
	//For GPU
	int offset_imaging_gpu_kernel_type;
	
	////For one way operators
	bool bUseOnewayPropagator;
	string pdip,oper;
	//int ntpx; //taper points
	int nws;  //number of requencies.
	string csouFileName,crecFileName;
	int nxPadOneway,nyPadOneway;
  
  //I use it to simulate sparse sampling of the image gather.
  cubeData<float>* pImageMask; 
  string fn_imageMask;
public:
   FDModelingParams(): propagateParams(), bUse4thOrder(false), bTake2ndTimeDerivForRTM(false),offset_imaging_gpu_kernel_type(0),
	  bUseOnewayPropagator(false),pdip("45"), oper("foufd"),nws(0),nxPadOneway(0),nyPadOneway(0),pImageMask(NULL)
	{}
};

//There will be one-way Wave Equation based Modeling Params.

inline void setFDwavfieldShape(hypercube& wavfieldShape, int stenWid, const hypercube& velGridShape) {
   int ndim = velGridShape.get_ndim();
   vector<axis> axes;
   axis ax;
   for (int i=1;i<=ndim;i++) //get_axis is 1 based.
   {
      ax = velGridShape.get_axis(i);
      ax.n = ax.n+2*stenWid;
	  ax.o = ax.o - stenWid*ax.d;
      axes.push_back(ax);
   }
   wavfieldShape.set_axes(axes);
}

};
#endif
