#ifndef __SOLVERNONLINEARSEARCH_H_INCL__
#define __SOLVERNONLINEARSEARCH_H_INCL__

#include "stdafx.h"
#include "dmRsfDump.h"
#include "cubeData.hpp"
#include "gVec.h"
#include "objFunc.h"
#include <boost/shared_ptr.hpp>

namespace Yang {

//Free the ptr memories in a vector
template <typename T>
void deleteVecPointers(vector<T*>& ptrVec) {
  for (size_t i=0; i<ptrVec.size(); i++) {
    delete(ptrVec[i]);
  }
}

//Dump the itermediate models
inline void dumpVecModels(vector<gVecComponent*>& modelptrVec, const string& tag) {
  size_t vecSize = modelptrVec.size();
  if (vecSize>0) {
    dmRsf_ocDump<float> dumpModels(tag,tag);
    dumpModels.init(&(modelptrVec[0]->domain()));
    for (size_t i=0; i<vecSize; i++) {
      dumpModels.dumpFrame(*modelptrVec[i]);
    }
  }
}

/** 
 * solver non-linear, perform a line search along the update 
 * direction 
 */

class smoother_intf;

struct solverParam {
  //For clipping the model
  float fmaxValAllow, fminValAllow;
  
  int maxIter;
  float errTol; //error tolerance for termination of iteration
  float fMaxPerturbScale;
  bool bDecreaseMaxStepLengthWithIteration;
  //The formula is stepLenMax = (base1 + (1-base1)*(fdecay^(iter)))*fMaxPerturbScale
  float fbase1,fdecay;
  
  typedef enum tagSmoothType {
    sm_spline,
    sm_rect,
    sm_alongDip
  } smoothType;
  smoothType eSmooth;
  
  bool bSmoothSearchDir,bChangeSmoothParamWithIteration;
  int   nSpl[3]; //number of points per axis
  float fsSpl[3]; //scale nxspl by these numbers every iteration
  //From coarse grid to finer grid.
  //int rect[3]; 
  int repeat;
  
  bool      bMaskGradient; //if mask gradient
  cubeDataF gradMask;
  
  solverParam(): fmaxValAllow(0.0f), fminValAllow(0.0f),
    maxIter(0), errTol(0.0f), fMaxPerturbScale(0.01f), 
    bDecreaseMaxStepLengthWithIteration(false), fbase1(1.f),fdecay(0.9f),
    bSmoothSearchDir(true),bChangeSmoothParamWithIteration(true) {
    nSpl[0] =nSpl[1]  =nSpl[2]  =0;
    fsSpl[0]=fsSpl[1] =fsSpl[2] =0.0f;
    eSmooth = sm_spline; 
  }
  
};

class solverNonlinearSearch {
public:
    objFunc* m_pObjFunc;
    smoother_intf* m_pSmoother;
    solverParam* m_pSolverPar;
protected:
   float m_curObjFuncValue; //Prev Obj Function Value
   float m_RMS_initModel; //Initial Model's Root-mean-square, used as a scaling reference
   
   auto_ptr<gVecComponent> m_apCurModel;
   auto_ptr<gVecComponent> m_apTrialModel; //A temp variable used in updateModel;
   
public:
	// Constructor
	solverNonlinearSearch(objFunc* pObj, const gVecComponent* pInitModel, solverParam* pSolParam);

	// Destructor
	virtual ~solverNonlinearSearch();

	
	//update the model given a search direction
	bool updateModel( const gVecComponent& searchDir, float maxPerturbScale);
	
	const gVecComponent& getCurModel() const {
	   return *m_apCurModel;
	}
   
	void solve( gVecComponent& finalModel, stdVecFloat& finalErrorVec, int& finalIter,
              std::vector<gVecComponent*>* pVecModels = NULL,
              std::vector<gVecComponent*>* pVecGrads = NULL,
              vector<gVecComponent*>* pVecDatas = NULL, bool bDataInCore = true);

  virtual void smoothGradient(const gVecComponent& inputSearchDir, gVecComponent& searchDir);
protected:
  void setSmoothParam(int nSpl[3]);
  void storeDataVec(int ind);
  void loadDataVec(int ind);
    
  stdVecInt mCur_nspl; //Spline nodes amount in current iterations.

  
  cubeData<float> calDataCubes[3];
  string          calDataFns[3]; 
  bool bDataInCore;
    
private:
	// Copy constructor
	// TODO: Uncomment the copy constructor when you need it.
	solverNonlinearSearch(const solverNonlinearSearch& src);

	// Assignment operator
	// TODO: Uncomment the assignment operator when you need it.
	solverNonlinearSearch& operator=(const solverNonlinearSearch& src);
};

// TODO: Uncomment the copy constructor when you need it.
//inline solverNonlinearSearch::solverNonlinearSearch(const solverNonlinearSearch& src)
//{
//   // TODO: copy
//}

// TODO: Uncomment the assignment operator when you need it.
//inline solverNonlinearSearch& solverNonlinearSearch::operator=(const solverNonlinearSearch& rhs)
//{
//   if (this == &rhs) {
//      return *this;
//   }
//
//   // TODO: assignment
//
//   return *this;
//}

}

#endif // __SOLVERNONLINEARSEARCH_H_INCL__
